from typing import Union

import numpy as np
import pandas as pd
from ripser import Rips, ripser
from scipy import sparse
from scipy.spatial.distance import pdist, squareform


class TopologicalTransformation:
    """Decomposes the given time series with a singular-spectrum analysis. Assumes the values of the time series are
    recorded at equal intervals.

    Args:
        time_series: Time series to be decomposed.
        max_simplex_dim: Maximum dimension of the simplices to be used in the Rips filtration.
        epsilon: Maximum distance between two points to be considered connected by an edge in the Rips filtration.
        persistence_params: ...
        window_length: Length of the window to be used in the rolling window function.

    Attributes:
        epsilon_range (np.ndarray): Range of epsilon values to be used in the Rips filtration.

    """

    def __init__(self,
                 time_series: Union[pd.Series, np.ndarray, list] = None,
                 max_simplex_dim: int = None,
                 epsilon: int = None,
                 persistence_params: dict = None,
                 window_length: int = None):
        self.time_series = time_series
        self.time_series = np.array(self.time_series)
        self.time_series = self.time_series.astype(float)
        self.max_simplex_dim = max_simplex_dim
        self.epsilon_range = self.__create_epsilon_range(epsilon)
        self.persistence_params = persistence_params

        if self.persistence_params is None:
            self.persistence_params = {
                'coeff': 2,
                'do_cocycles': False,
                'verbose': False}

        self.__window_length = window_length

    @staticmethod
    def __create_epsilon_range(epsilon):
        return np.array([y * float(1 / epsilon) for y in range(epsilon)])

    @staticmethod
    def __compute_persistence_landscapes(ts):

        N = len(ts)
        I = np.arange(N - 1)
        J = np.arange(1, N)
        V = np.maximum(ts[0:-1], ts[1::])

        # Add vertex birth times along the diagonal of the distance matrix
        I = np.concatenate((I, np.arange(N)))
        J = np.concatenate((J, np.arange(N)))
        V = np.concatenate((V, ts))

        # Create the sparse distance matrix
        D = sparse.coo_matrix((V, (I, J)), shape=(N, N)).tocsr()
        dgm0 = ripser(D, maxdim=0, distance_matrix=True)['dgms'][0]
        dgm0 = dgm0[dgm0[:, 1] - dgm0[:, 0] > 1e-3, :]

        allgrid = np.unique(dgm0.flatten())
        allgrid = allgrid[allgrid < np.inf]

        xs = np.unique(dgm0[:, 0])
        ys = np.unique(dgm0[:, 1])
        ys = ys[ys < np.inf]

    @staticmethod
    def rolling_window(array: np.array, window: int) -> np.array:
        """Takes in an array and return array of rolling windows of specified length.

        Args:
            array: Array to be rolled.
            window: Length of the window.

        Returns:
            Array of rolling windows.

        """
        shape = array.shape[:-1] + (array.shape[-1] - window + 1, window)
        strides = array.strides + (array.strides[-1],)
        a_windowed = np.lib.stride_tricks.as_strided(array, shape=shape, strides=strides)
        return a_windowed

    def time_series_to_point_cloud(self, array: np.array = None,
                                   dimension_embed=2) -> np.array:
        """Convert a time series into a point cloud in the dimension specified by dimension_embed.

        Args:
            array: Time series to be converted.
            dimension_embed: dimension of Euclidean space in which to embed the time series into by taking
            windows of dimension_embed length, e.g. if the time series is ``[t_1,...,t_n]`` and dimension_embed
            is ``2``, then the point cloud would be ``[(t_0, t_1), (t_1, t_2),...,(t_(n-1), t_n)]``

        Returns:
            Collection of points embedded into Euclidean space of dimension = dimension_embed, constructed
            in the manner explained above.

        """

        assert len(self.time_series) >= dimension_embed, 'dimension_embed larger than length of time_series'

        if self.__window_length is None:
            self.__window_length = dimension_embed

        # compute point cloud
        if array is None:
            array = self.time_series

        point_cloud = self.rolling_window(array=array, window=dimension_embed)
        return np.array(point_cloud)

    def point_cloud_to_persistent_cohomology_ripser(self, point_cloud: np.array = None,
                                                    max_simplex_dim: int = None):

        # ensure epsilon_range is a numpy array
        epsilon_range = self.epsilon_range

        # build filtration
        self.persistence_params['maxdim'] = max_simplex_dim
        filtration = Rips(**self.persistence_params)

        if point_cloud is None:
            point_cloud = self.time_series_to_point_cloud()

        # initialize persistence diagrams
        diagrams = filtration.fit_transform(point_cloud)
        # Instantiate persistence landscape transformer
        # plot_diagrams(diagrams)

        # normalize epsilon distance in diagrams so max is 1
        diagrams = [np.array([dg for dg in diag if np.isfinite(dg).all()]) for diag in diagrams]
        diagrams = diagrams / max(
            [np.array([dg for dg in diag if np.isfinite(dg).all()]).max() for diag in diagrams if diag.shape[0] > 0])

        ep_ran_len = len(epsilon_range)

        homology = {dimension: np.zeros(ep_ran_len).tolist() for dimension in range(max_simplex_dim + 1)}

        for dimension, diagram in enumerate(diagrams):
            if dimension <= max_simplex_dim and len(diagram) > 0:
                homology[dimension] = np.array(
                    [np.array(((epsilon_range >= point[0]) & (epsilon_range <= point[1])).astype(int))
                     for point in diagram
                     ]).sum(axis=0).tolist()

        return homology

    def time_series_to_persistent_cohomology_ripser(self,
                                                    time_series: np.array,
                                                    max_simplex_dim: int) -> dict:
        """Wrapper function that takes in a time series and outputs the persistent homology object, along with other
        auxiliary objects.

        Args:
            time_series: Time series to be converted.
            max_simplex_dim: Maximum dimension of the simplicial complex to be constructed.

        Returns:
            Persistent homology object. Dictionary with keys in ``range(max_simplex_dim)`` and, the value ``hom[i]``
            is an array of length equal to ``len(epsilon_range)`` containing the betti numbers of the ``i-th`` homology
            groups for the Rips filtration.

        """

        homology = self.point_cloud_to_persistent_cohomology_ripser(point_cloud=time_series,
                                                                    max_simplex_dim=max_simplex_dim)
        return homology

    def time_series_rolling_betti_ripser(self, ts):

        point_cloud = self.rolling_window(array=ts, window=self.__window_length)
        homology = self.time_series_to_persistent_cohomology_ripser(point_cloud,
                                                                    max_simplex_dim=self.max_simplex_dim)
        df_features = pd.DataFrame(data=homology)
        cols = ["Betti_{}".format(i) for i in range(df_features.shape[1])]
        df_features.columns = cols
        df_features['Betti_sum'] = df_features.sum(axis=1)
        return df_features


class TSTransformer:
    def __init__(self, time_series, min_signal_ratio, max_signal_ratio, rec_metric):
        self.time_series = time_series
        self.recurrence_matrix = None
        self.threshold_baseline = [1, 5, 10, 15, 20, 25, 30]
        self.min_signal_ratio = min_signal_ratio
        self.max_signal_ratio = max_signal_ratio
        self.rec_metric = rec_metric

    def ts_to_recurrence_matrix(self,
                                eps=0.10,
                                steps=None):
        distance_matrix = pdist(metric=self.rec_metric, X=self.time_series[:, None])
        distance_matrix = np.floor(distance_matrix / eps)
        distance_matrix, steps = self.binarization(distance_matrix, threshold=steps)
        distance_matrix[distance_matrix > steps] = steps
        self.recurrence_matrix = squareform(distance_matrix)
        return self.recurrence_matrix

    def binarization(self, distance_matrix, threshold):
        best_threshold_flag = False
        signal_ratio_list = []
        if threshold is None:
            for threshold_baseline in self.threshold_baseline:
                threshold = threshold_baseline
                tmp_array = np.copy(distance_matrix)
                tmp_array[tmp_array < threshold_baseline] = 0.0
                tmp_array[tmp_array >= threshold_baseline] = 1.0
                signal_ratio = np.where(tmp_array == 0)[0].shape[0] / tmp_array.shape[0]

                if self.min_signal_ratio < signal_ratio < self.max_signal_ratio:
                    best_ratio = signal_ratio
                    distance_matrix = tmp_array
                    best_threshold_flag = True
                    if signal_ratio > best_ratio:
                        distance_matrix = tmp_array
                else:
                    signal_ratio_list.append(abs(self.max_signal_ratio - signal_ratio))

                del tmp_array

        if not best_threshold_flag:
            threshold = self.threshold_baseline[signal_ratio_list.index(min(signal_ratio_list))]
            distance_matrix[distance_matrix < threshold] = 0.0
            distance_matrix[distance_matrix >= threshold] = 1.0
        return distance_matrix, threshold

    def get_recurrence_metrics(self):
        if self.recurrence_matrix is None:
            return self.ts_to_recurrence_matrix()



import numpy as np
from statistics import mean
from scipy.signal import savgol_filter
from fedot_ind.core.models.detection.utils.get_time import get_current_time


def NormalizeData(ts: list, return_numpy: bool = False):
    """
    Normalization method for time seroes. 
    Output time series' values are between 0.0 and 1.0

    Args:
        ts (list): time series, ould be list or numpy.array
        return_numpy (bool, optional): If set to True, method result will be in 
        Numply.array format. Defaults to False.

    Raises:
        ValueError: If ts isn't list or Numpy.array

    Returns:
        list or Numpy.array: Normilized time series
    """

    #norm_1 = (ts - np.min(ts)) / (np.max(ts) - np.min(ts))
    if not isinstance(ts, list):
        #print(f"{get_current_time()} Math utils - Normalization: In normalization method not a list was sent. Possible np.array. Attempting to convert...")
        try:
            ts = ts.tolist()
            #print(f"{get_current_time()} Math utils - Normalization: Succesfull converted!")
        except:
            raise ValueError(F"Math utils - Normalization: An unexpected type of time series has been meet!")

    norm_list: list = list()
    min_value = min(ts)
    distance = max(ts) - min_value
    if distance == 0: 
        print(f"---[{get_current_time()}] Math utils - Normalization: Distance between min and max is zero! Ts without changes!")
        norm_list = [0] * len(ts)
    else:
        for value in ts:
            norm_list.append((value - min_value) / distance)
        if return_numpy:
            norm_list = np.array(norm_list)

    return norm_list

def NormalizeData_1(data):
    from sklearn.preprocessing import minmax_scale
    data = minmax_scale(data, feature_range=(0,1))
    return data

def NormalizeDataForDetectors(data):
    data = np.append(data, [0])
    output = (data - np.min(data)) / (np.max(data) - np.min(data))
    output = output[:-1]
    #output = [float(i)/sum(data) for i in data]
    return output

def Cut_data(data: list, threshold: float):
    """
    This method cut data - everython under threshold will be put to zero

    Args:
        data (list): list
        threshold (float): float

    Returns:
        out_data (list): cutted data
    """
    #values = []
    #values_dict = {}
    #for value in data:
    #    if not value in values_dict: 
    #        values.append(value)
    #        values_dict[value] = 0
    #values.sort()
    #length = len(values)
    if not 0 <= threshold <= 1:
        threshold = 0.99
    #q = int(length * threshold)
    #if q >= len(values): q -= 1
    #quantile = values[int(length * threshold)]
    quantile = threshold
    if data[0] < 0:  
        out_data = []
        for element in data:
            if element >= quantile: out_data.append(0)
            else: out_data.append(element)
    else:
        out_data = []
        for element in data:
            if element <= quantile: out_data.append(0)
            else: out_data.append(element)
    
    return out_data

def clean_begin_of_ts(ts: list, start_part_len: int, desirable_value: float = 0) -> list:
    """
    Because of very big nouces at the start and the end of time series 
    we have to clean it - made it 0 or another value

    Args:
        ts (list): some time serie
        start_part_len (int): len of start part. Method will change all
            elements of ts from 0 to <start_part_len>
        desirable_value (float): desirable value of elements in 
            this part, 0 by default
    Returns:
        list: changed time serie
    """
    mult = 5
    if len(ts) > start_part_len:
        for j in range(start_part_len):
            ts[j] = desirable_value
        if len(ts) < start_part_len * mult:
            print(f"---[{get_current_time()}] Math utils - Begin cleaner warning: Time series is short. <{start_part_len} * {mult}! Anomalies could be detected with errors or doesn't detected at all!")
            #warnings.warn(f"Transformation warning: Time series is short. <{start_part_len} * {mult}! Anomalies could be detected with errors or doesn't detected at all!")
    else:
        raise ValueError(f"Transformation error: Time series is too short. <{start_part_len}!")
    return ts

def clean_end_of_ts(ts: list, end_part_len: int, desirable_value: float = 0) -> list:
    """
    Because of very big nouces at the start and the end of time series 
    we have to clean it - made it 0 or another value

    Args:
        ts (list): some time serie
        end_part_len (int): len of start part. Method will change all
            elements of ts from <start_part_len> to end of ts
        desirable_value (float): desirable value of elements in 
            this part, 0 by default
    Returns:
        list: changed time serie
    """
    mult = 5
    if len(ts) > end_part_len:
        for j in range(len(ts)-end_part_len, len(ts)):
            ts[j] = desirable_value
        if len(ts) < end_part_len * mult:
            print(f"---[{get_current_time()}] Math utils - End cleaner warning: Time series is short. <{end_part_len} * {mult}! Anomalies could be detected with errors or doesn't detected at all!")
            #warnings.warn(f"Transformation warning: Time series is short. <{end_part_len} * {mult}! Anomalies could be detected with errors or doesn't detected at all!")
    else:
        raise ValueError(f"Transformation error: Time series is too short. <{end_part_len}!")
    return ts

def move_ts_to_desirable_average_level(ts: list, desirable_average: float = 0) -> list:
    """
    This method moves time series up or down along Y-axis
    to the level when average of this time series became equial to desirable

    Args:
        ts (list): list of float numbers(possible any number values)
        desirable_average (float, optional): Desirable level of average. Defaults to 0.

    Returns:
        list: changed ts
    """
    if mean(ts) < desirable_average:
        mean_distance = desirable_average - mean(ts)
        for j in range(len(ts)):
            ts[j] = ts[j] + abs(mean_distance)
    else:
        mean_distance = desirable_average - mean(ts)
        for j in range(len(ts)):
            ts[j] = ts[j] - abs(mean_distance)
    return ts

def smooth_ts_for_transforming(
    ts: list, 
    cycles_count: int, 
    window_len_of_first_filer: int = 87,
    window_len_of_second_filer: int = 31) -> list:
    """
    Method smooth time series.

    Args:
        ts (list): time serie
        cycles_count (int): count of cycles of smoothing
        window_len_of_first_filer (int, optional): An odd int value. Magic len of window for first of two smoother methods. Defaults to 87.
        window_len_of_second_filer (int, optional): An odd int value. Magic len of window for second of two smoother methods. Defaults to 31.

    Returns:
        list: smoothed ts
    """
    
    for _ in range(cycles_count):
        ts = savgol_filter(ts, window_len_of_first_filer, 1) 
        ts = savgol_filter(ts, window_len_of_second_filer, 1)
    return ts


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error

from fedot_ind.core.metrics.loss.basis_loss import basis_approximation_metric
from fedot_ind.core.operation.transformation.basis.abstract_basis import BasisDecompositionImplementation


class ChebyshevBasis(BasisDecompositionImplementation):

    def _get_basis(self, data, n_components: int = None):
        self.basis = []
        self.coefs = []
        self.rmse = []
        time_domain = np.arange(self.data_range)
        value_domain = np.arange(np.min(data), np.min(data), 0.1, dtype=float)
        for i in range(1, self.n_components + 1):
            # poly_coefs = np.polynomial.chebyshev.chebgrid2d(x=time_domain, y=value_domain, c=i)
            poly_coefs = np.polynomial.chebyshev.chebfit(x=time_domain, y=data, deg=i)
            # poly_coefs = np.ones(shape=len(time_domain))
            new_y = np.polynomial.chebyshev.chebval(time_domain, poly_coefs)
            self.coefs.append(poly_coefs)
            self.basis.append(new_y)
            self.rmse.append(mean_squared_error(y_true=data,
                                                y_pred=new_y,
                                                squared=False))

    def _transform(self, features):
        self._get_basis(features, self.n_components)
        self.basis = np.array(self.basis).T
        self.data = features
        # coef =Ridge(alpha=0.5).fit(self.basis, self.data.T)
        return self.basis

    def evaluate_derivative(self, order):
        domain = np.arange(self.data_range)
        self.derv_coef = []
        for poly_coef in self.coefs:
            derivative_of_poly = np.polynomial.legendre.legder(poly_coef, order)
            new_y = np.polynomial.legendre.legval(domain, derivative_of_poly)
            self.derv_coef.append(new_y)

    def analytical_form(self):
        polynom_number = basis_approximation_metric(metric_values=np.array(self.rmse),
                                                    derivation_coef=self.derv_coef,
                                                    regularization_coef=0.9)
        polynom_coef = self.coefs[polynom_number]
        p = np.polynomial.chebyshev.cheb2poly(polynom_coef)
        anylytical_form = ''
        for index, coef in enumerate(p):
            degree = f' +{round(coef, 3)}*x^({len(p) - index - 1}) '
            anylytical_form = anylytical_form + degree
        anylytical_form = anylytical_form.replace('+-', '-')
        print(f'Best degree of Chebyshev Polynomial is - {polynom_number + 1}')
        print(anylytical_form)
        return polynom_number

    def show(self, visualisation_type: str = 'basis representation', **kwargs):
        if visualisation_type == 'basis representation':
            A = np.array(self.basis)
        else:
            A = np.array(self.derv_coef)
        A = pd.DataFrame(A)
        A['original'] = self.data
        A[[kwargs['basis_function'], 'original']].plot()
        plt.show()



import torch
import torch.nn as nn


def correct_sizes(sizes):
    corrected_sizes = [s if s % 2 != 0 else s - 1 for s in sizes]
    return corrected_sizes


def pass_through(X):
    return X


class Flatten(nn.Module):
    def __init__(self, out_features):
        super(Flatten, self).__init__()
        self.output_dim = out_features

    def forward(self, x):
        return x.view(-1, self.output_dim)


class Reshape(nn.Module):
    def __init__(self, out_shape):
        super(Reshape, self).__init__()
        self.out_shape = out_shape

    def forward(self, x):
        return x.view(-1, *self.out_shape)
